#include <vector>
#include <algorithm>
#include <utility>

#include "SRTree.h"

SpatialIndex::RTree::SRTree::SRTree(RTree* _tree, uint _disc_factor, int _index_type, uint param) {
  disc_factor = _disc_factor;
  numObj = _tree->m_stats.getNumberOfData();
  numEntriesPerNode = _tree->m_indexCapacity;
  index_type = _index_type;
  
  SpatialIndex::RTree::NodePtr cN = _tree->readNode(_tree->m_rootID);
  
  lower0 = (int)round(cN->m_nodeMBR.m_pLow[0]*disc_factor);
  lower1 = (int)round(cN->m_nodeMBR.m_pLow[1]*disc_factor);

  root = createNode(_tree, cN, param);

}

void SpatialIndex::RTree::SRTree::range_query(double l0, double l1, double h0, double h1, long* result, int* resultSize) {
  root->range_query_disc(l0, l1, h0, h1, (int)round(l0*disc_factor), (int)round(l1*disc_factor), (int)round(h0*disc_factor), (int)round(h1*disc_factor), 
			 lower0, lower1, result, resultSize);
}

uint SpatialIndex::RTree::SRTree::memory_usage() {
  
  return 2*sizeof(int)+root->memory_usage();
  //return 20 * numEntriesPerNode * numObj / (numEntriesPerNode-1);

}

void SpatialIndex::RTree::SRTree::memory_usage(int &total, int &coords, int &ids, int &structure) {
  total = 2*sizeof(int);
  structure = total;
  coords = 0;
  ids = 0;
  root->memory_usage(total, coords, ids, structure);
}

bool myNodePtrCmp(pair <size_t, double> n1, pair <size_t, double> n2) {
  return n1.second < n2.second;
}

SNode* SpatialIndex::RTree::SRTree::createNode(RTree* _tree, NodePtr cNode, uint param) {
  SNode* nNode;
  
  if(index_type == SRTREE_TYPE_REG) {
    if(cNode->m_level != 0) {
      nNode = new SNodeInt(cNode->m_children);
      for (size_t cChild = 0; cChild < cNode->m_children; cChild++) {
	  Region cChildRegion = *(cNode->m_ptrMBR[cChild]);
	  nNode->addEntry(cChild, cChildRegion.m_pLow[0], cChildRegion.m_pLow[1], cChildRegion.m_pHigh[0], cChildRegion.m_pHigh[1],
			  (size_t)createNode(_tree, _tree->readNode(cNode->m_pIdentifier[cChild]), param), disc_factor, 0, 0);
      }  
    } else {
      nNode = new SNodeLeaf(cNode->m_children);
      for (size_t cChild = 0; cChild < cNode->m_children; cChild++) {
	  Region cChildRegion = *(cNode->m_ptrMBR[cChild]);
	  nNode->addEntry(cChild, cChildRegion.m_pLow[0], cChildRegion.m_pLow[1], cChildRegion.m_pHigh[0], cChildRegion.m_pHigh[1],
			  (size_t)cNode->m_pIdentifier[cChild], disc_factor, 0, 0);
      }
    }
  } else if(index_type == SRTREE_TYPE_COMP) {
    vector<pair <size_t, double> > orderedChildren;
    for (size_t cChild = 0; cChild < cNode->m_children; cChild++) {
      orderedChildren.push_back(make_pair(cChild, cNode->m_ptrMBR[cChild]->m_pLow[0]));
    }
    sort(orderedChildren.begin(), orderedChildren.end(), myNodePtrCmp);
    vector<pair <size_t, double> >::iterator it;
    double pl0 = cNode->m_nodeMBR.m_pLow[0];
    double pl1 = cNode->m_nodeMBR.m_pLow[1];
    int cEntry = 0;
    if(cNode->m_level != 0) {
      nNode = new SNodeIntComp(cNode->m_children);
      for (it=orderedChildren.begin() ; it < orderedChildren.end(); it++ ) {
	size_t cChild = it->first;
	Region cChildRegion = *(cNode->m_ptrMBR[cChild]);
	nNode->addEntry(cEntry++, cChildRegion.m_pLow[0], cChildRegion.m_pLow[1], cChildRegion.m_pHigh[0], cChildRegion.m_pHigh[1],
			  (size_t)createNode(_tree, _tree->readNode(cNode->m_pIdentifier[cChild]), param), disc_factor, pl0, pl1);
	pl0 = cChildRegion.m_pLow[0];
      }
    } else {
      nNode = new SNodeLeafComp(cNode->m_children);
      for (it=orderedChildren.begin() ; it < orderedChildren.end(); it++ ) {
	size_t cChild = it->first;
	Region cChildRegion = *(cNode->m_ptrMBR[cChild]);
	nNode->addEntry(cEntry++, cChildRegion.m_pLow[0], cChildRegion.m_pLow[1], cChildRegion.m_pHigh[0], cChildRegion.m_pHigh[1],
			  (size_t)cNode->m_pIdentifier[cChild], disc_factor, pl0, pl1);
	pl0 = cChildRegion.m_pLow[0];
      }
    }
  /*} else if(index_type == SRTREE_TYPE_HYB) {
    if(cNode->m_level == 0) {
	nNode = new SNodeLeaf(cNode->m_children);
	for (size_t cChild = 0; cChild < cNode->m_children; cChild++) {
	    Region cChildRegion = *(cNode->m_ptrMBR[cChild]);
	    nNode->addEntry(cChild, cChildRegion.m_pLow[0], cChildRegion.m_pLow[1], cChildRegion.m_pHigh[0], cChildRegion.m_pHigh[1],
			    (size_t)cNode->m_pIdentifier[cChild], disc_factor, 0, 0);
	}
    } else {
      if(param != 0) {
	nNode = new SNodeInt(cNode->m_children);
	for (size_t cChild = 0; cChild < cNode->m_children; cChild++) {
	    Region cChildRegion = *(cNode->m_ptrMBR[cChild]);
	    nNode->addEntry(cChild, cChildRegion.m_pLow[0], cChildRegion.m_pLow[1], cChildRegion.m_pHigh[0], cChildRegion.m_pHigh[1],
			    (size_t)createNode(_tree, _tree->readNode(cNode->m_pIdentifier[cChild]), param-1), disc_factor, 0, 0);
	}
      } else {
	vector<INPUT_MBR> allMBRs;
	collectMBRs(_tree, cNode, allMBRs);
	INPUT_MBR* argMBRs = (INPUT_MBR*)malloc(allMBRs.size()*sizeof(INPUT_MBR));
	for(int i = 0; i < allMBRs.size(); i++) {
	  argMBRs[i] = allMBRs.at(i);
	}
	nNode = new SNodeRDT(cNode->m_children, disc_factor, argMBRs, allMBRs.size());
      }
    }
    */
  } else {
    return NULL;
  }
  
  return nNode;
}

void SpatialIndex::RTree::SRTree::collectMBRs(RTree* _tree, NodePtr cNode, vector <INPUT_MBR> &collector) {
    if(cNode->m_level != 0) {
      for (size_t cChild = 0; cChild < cNode->m_children; cChild++) {
	collectMBRs(_tree, _tree->readNode(cNode->m_pIdentifier[cChild]), collector);
      }  
    } else {
      for (size_t cChild = 0; cChild < cNode->m_children; cChild++) {
	  Region cChildRegion = *(cNode->m_ptrMBR[cChild]);
	  INPUT_MBR tmpMBR;
	  tmpMBR.id = cNode->m_pIdentifier[cChild];
	  tmpMBR.mbr.point1.lon = cChildRegion.m_pLow[0];
	  tmpMBR.mbr.point1.lat = cChildRegion.m_pLow[1];
	  tmpMBR.mbr.point2.lon = cChildRegion.m_pHigh[0];
	  tmpMBR.mbr.point2.lat = cChildRegion.m_pHigh[1];
	  collector.push_back(tmpMBR);
      }
    }  
}
